/*
 * NDPluginFFT.cpp
 *
 * FFT plugin
 * Author: Mark Rivers
 *
 * Created January 29, 2015
 *
 */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include <iocsh.h>

#include <asynDriver.h>

#include <epicsExport.h>
#include "NDPluginDriver.h"
#include "NDPluginFFT.h"

extern "C" {
void fourn(float data[], unsigned long nn[], int ndim, int isign);
}

static const char *driverName = "NDPluginFFT";

template <typename epicsType>
asynStatus NDPluginFFT::computeFFT(NDArray *pArray)
{
  unsigned long ldims[2];
  size_t dims[2];
  size_t sizeX, sizeY;
  size_t i, j;
  float *FFT;
  int FFTDirection;
  double scaleFactor;
  int outputDataType;
  NDArray *pTemp, *pArrayOut;
  epicsType *pDataIn = (epicsType *)pArray->pData;
  epicsType *pIn;
  float *pDataOut;
  float *pOut;
  NDArrayInfo arrayInfo;

  pArray->getInfo(&arrayInfo);
  getIntegerParam(P_Direction_,   &FFTDirection);
  getDoubleParam (P_ScaleFactor_, &scaleFactor);
  getIntegerParam(P_DataType_,    &outputDataType);
  if (outputDataType == -1) outputDataType = pArray->dataType;

  /* Release the lock; The code after this is computationally intensive and does not access any shared data */
  this->unlock();

  // The fourn function only works on dimensions that are powers of 2 so find the first power of
  // 2 that is >= input dimension
  for (sizeX=1; sizeX < arrayInfo.xSize; sizeX*=2);
  for (sizeY=1; sizeY < arrayInfo.ySize; sizeY*=2);
  // Create float array which has 2 * sizeX * sizeY, because FFT is done in place and output is complex
  FFT = (float *)calloc(2*sizeX*sizeY, sizeof(float));
  for (i=0; i<arrayInfo.ySize; i++) {
    pIn  = pDataIn  + i*arrayInfo.xSize;
    pOut = FFT + 2*i*sizeX;
    for (j=0; j<arrayInfo.xSize; j++) {
        pOut[2*j] = (float) pIn[j];
    }
  }
  // Take the FFT of the array
  ldims[0] = sizeX;
  ldims[1] = sizeY;
  // fourn wants a pointer to one element before the first one for the data and dimensions arrays
  fourn(FFT-1, ldims-1, 2, FFTDirection);
  
  // Allocate an NDArray of type NDFloat32
  dims[0] = sizeX;
  dims[1] = sizeY;
  pTemp = this->pNDArrayPool->alloc(2, dims, NDFloat32, 0, NULL);
  if (FFTDirection == 1) scaleFactor = scaleFactor / (dims[0] * dims[1]);
  pDataOut = (float *)pTemp->pData;
  for (i=0; i<arrayInfo.nElements; i++) {
    pDataOut[i] = sqrt(FFT[2*i]*FFT[2*i] + FFT[2*i+1]*FFT[2*i+1]) * scaleFactor;
  }
  
  // Convert the FFT to the desired output data type
  this->pNDArrayPool->convert(pTemp, &pArrayOut, (NDDataType_t)outputDataType);
  free(FFT);
  pTemp->release();
  pArrayOut->uniqueId = pArray->uniqueId;
  pArrayOut->timeStamp = pArray->timeStamp;
  pArrayOut->epicsTS = pArray->epicsTS;
  pArrayOut->pAttributeList->clear();
  pArray->pAttributeList->copy(pArrayOut->pAttributeList);
  pArrays[0] = pArrayOut;
  // Take the lock again, we are done with the computationally intensive task
  this->lock();
  return(asynSuccess);
}

/** Callback function that is called by the NDArray driver with new NDArray data.
  * Grabs the current NDArray and applies the selected FFTs to the data.  Apply the FFTs in order.
  * \param[in] pArray  The NDArray from the callback.
  */
void NDPluginFFT::processCallbacks(NDArray *pArray){
  
  asynStatus status;
  static const char* functionName = "processCallbacks";

  /* Call the base class method */
  NDPluginDriver::processCallbacks(pArray);

  if (pArray->ndims != 2) {
    asynPrint( this->pasynUserSelf, ASYN_TRACE_ERROR, 
        "%s::%s, this method is meant to FFT 2Dimages when the number of dimensions is <= 3\n",
        pluginName, functionName);
    return;
  }
  
  /* Previous version of the array was held in memory.  Release it. */
  if (this->pArrays[0]) {
    this->pArrays[0]->release();
    this->pArrays[0] = NULL;
  }

  switch(pArray->dataType) {
      case NDInt8:
          status = computeFFT<epicsInt8>(pArray);
          break;
      case NDUInt8:
          status = computeFFT<epicsUInt8>(pArray);
          break;
      case NDInt16:
          status = computeFFT<epicsInt16>(pArray);
          break;
      case NDUInt16:
          status = computeFFT<epicsUInt16>(pArray);
          break;
      case NDInt32:
          status = computeFFT<epicsInt32>(pArray);
          break;
      case NDUInt32:
          status = computeFFT<epicsUInt32>(pArray);
          break;
      case NDFloat32:
          status = computeFFT<epicsFloat32>(pArray);
          break;
      case NDFloat64:
          status = computeFFT<epicsFloat64>(pArray);
          break;
      default:
          status = asynError;
      break;
  }
  if (status) {
    asynPrint(pasynUserSelf, ASYN_TRACE_ERROR,
      "%s::%s error calling computeFFT, status=%d\n", 
      driverName, functionName, status);
  }
  doCallbacksGenericPointer(pArrays[0], NDArrayData, 0);
  callParamCallbacks();
}


/** Constructor for NDPluginFFT; most parameters are simply passed to NDPluginDriver::NDPluginDriver.
  * After calling the base class constructor this method sets reasonable default values for all of the
  * FFT parameters.
  * \param[in] portName The name of the asyn port driver to be created.
  * \param[in] queueSize The number of NDArrays that the input queue for this plugin can hold when
  *      NDPluginDriverBlockingCallbacks=0.  Larger queues can decrease the number of dropped arrays,
  *      at the expense of more NDArray buffers being allocated from the underlying driver's NDArrayPool.
  * \param[in] blockingCallbacks Initial setting for the NDPluginDriverBlockingCallbacks flag.
  *      0=callbacks are queued and executed by the callback thread; 1 callbacks execute in the thread
  *      of the driver doing the callbacks.
  * \param[in] NDArrayPort Name of asyn port driver for initial source of NDArray callbacks.
  * \param[in] NDArrayAddr asyn port driver address for initial source of NDArray callbacks.
  * \param[in] maxBuffers The maximum number of NDArray buffers that the NDArrayPool for this driver is
  *      allowed to allocate. Set this to -1 to allow an unlimited number of buffers.
  * \param[in] maxMemory The maximum amount of memory that the NDArrayPool for this driver is
  *      allowed to allocate. Set this to -1 to allow an unlimited amount of memory.
  * \param[in] priority The thread priority for the asyn port driver thread if ASYN_CANBLOCK is set in asynFlags.
  * \param[in] stackSize The stack size for the asyn port driver thread if ASYN_CANBLOCK is set in asynFlags.
  */
NDPluginFFT::NDPluginFFT(const char *portName, int queueSize, int blockingCallbacks,
             const char *NDArrayPort, int NDArrayAddr, int maxBuffers, size_t maxMemory,
             int priority, int stackSize)
  /* Invoke the base class constructor */
  : NDPluginDriver(portName, queueSize, blockingCallbacks,
                   NDArrayPort, NDArrayAddr, 1, NUM_FFT_PARAMS, maxBuffers, maxMemory,
                   asynInt32ArrayMask | asynFloat64ArrayMask | asynGenericPointerMask,
                   asynInt32ArrayMask | asynFloat64ArrayMask | asynGenericPointerMask,
                   ASYN_MULTIDEVICE, 1, priority, stackSize)
{
  //static const char *functionName = "NDPluginFFT";

  createParam(NDPluginFFTScaleFactorString, asynParamFloat64, &P_ScaleFactor_);
  createParam(NDPluginFFTDataTypeString,      asynParamInt32, &P_DataType_);
  createParam(NDPluginFFTDirectionString,     asynParamInt32, &P_Direction_);

  /* Set the plugin type string */
  setStringParam(NDPluginDriverPluginType, "NDPluginFFT");

  /* Try to connect to the array port */
  connectToArrayPort();
}

/** Configuration command */
extern "C" int NDFFTConfigure(const char *portName, int queueSize, int blockingCallbacks,
                                    const char *NDArrayPort, int NDArrayAddr,
                                    int maxBuffers, size_t maxMemory,
                                    int priority, int stackSize)
{
  new NDPluginFFT(portName, queueSize, blockingCallbacks, NDArrayPort, NDArrayAddr,
              maxBuffers, maxMemory, priority, stackSize);
  return(asynSuccess);
}

/* EPICS iocsh shell commands */
static const iocshArg initArg0 = { "portName",iocshArgString};
static const iocshArg initArg1 = { "frame queue size",iocshArgInt};
static const iocshArg initArg2 = { "blocking callbacks",iocshArgInt};
static const iocshArg initArg3 = { "NDArrayPort",iocshArgString};
static const iocshArg initArg4 = { "NDArrayAddr",iocshArgInt};
static const iocshArg initArg5 = { "maxBuffers",iocshArgInt};
static const iocshArg initArg6 = { "maxMemory",iocshArgInt};
static const iocshArg initArg7 = { "priority",iocshArgInt};
static const iocshArg initArg8 = { "stackSize",iocshArgInt};
static const iocshArg * const initArgs[] = {&initArg0,
                                            &initArg1,
                                            &initArg2,
                                            &initArg3,
                                            &initArg4,
                                            &initArg5,
                                            &initArg6,
                                            &initArg7,
                                            &initArg8};
static const iocshFuncDef initFuncDef = {"NDFFTConfigure",9,initArgs};
static void initCallFunc(const iocshArgBuf *args)
{
  NDFFTConfigure(args[0].sval, args[1].ival, args[2].ival,
                       args[3].sval, args[4].ival, args[5].ival,
                       args[6].ival, args[7].ival, args[8].ival);
}

extern "C" void NDFFTRegister(void)
{
  iocshRegister(&initFuncDef,initCallFunc);
}

extern "C" {
epicsExportRegistrar(NDFFTRegister);
}
